#include "env.hpp"


double env::evaluate(team *m,
		     point *p)
{  
  double rewardSum = 0;
  long action;
  
  vector < double > state;
  bool end = false;

  /* Get initial state defined by point. */
  p->state(state);

  while(end == false)
    {
      action = m->getAction(state);
      rewardSum += act(p, action, state, end);
    }

#ifdef MYDEBUG
  cout << "scm::evaluateEnv " << m->id() << " on " << p->id();
  cout << " reward " << rewardSum << endl;
#endif
  
  return rewardSum;
}

long explicitEnv::getLabelIndex(long c)
{ 
  map < long, long >::iterator iter = _labelMap.find(c); 
  
  if(iter == _labelMap.end()) 
    die(__FILE__, __FUNCTION__, __LINE__, "bad label"); 
  
  return (*iter).second; 
} 

point * explicitEnv::initUniformPoint(long gtime, 
				      set < long > &usedIds)
{
  long index;
  set < long > :: iterator iditerend = usedIds.end();
  
  do
    {
      index = (long) (drand48() * size());
    }
  while(usedIds.find(index) != iditerend);
  
  return new point(index, _dim, gtime, _pointSet[index]->first, _pointSet[index]->second);  
}

point * explicitEnv::initUniformPointLabel(long gtime, 
					   set < long > &usedIds,
					   long label)
{
  /* Of the new point in _pointSet (also its id). */
  long index;

  set < long > :: iterator iditerend = usedIds.end();

  long tries = 0;
  
  /* Retrieve appropriate index vector from _index. */
  vector < long > *indexvec = _index[getLabelIndex(label)];
  long size = indexvec->size();
  
  do
    {
      /* Get random point of label 'label'. */
      index = (*indexvec)[(long) (drand48() * size)];      

      if(tries++ > _maxTries) /* Cannot find a point of that label. */
	return 0;
    }
  while(usedIds.find(index) != iditerend);

  return new point(index, _dim, gtime, _pointSet[index]->first, _pointSet[index]->second);  

//   long index;
//   set < long > :: iterator iditerend = usedIds.end();
//   long tries = 0;
  
//   do
//     {
//       index = (long) (drand48() * size());      
//       if(tries++ > _maxTries) /* Cannot find the point of that label. */
// 	return 0;
//     }
//   while((_pointSet[index]->second != label) || (usedIds.find(index) != iditerend));
  
//   return new point(index, _dim, gtime, _pointSet[index]->first, _pointSet[index]->second);  
}

datasetEnv::datasetEnv(string file, 
		       int dim, 
		       long size)
  : explicitEnv(dim, file)
{
  set < long > labels;
  set < long > :: iterator laiter;
  
  map < long, long > labelCount;

  _pointSet.reserve(size);
  
  ifstream infile(file.c_str(), ios::in);
  
  if(infile == 0)
    die(__FILE__, __FUNCTION__, __LINE__, "cannot open dataset");
  
  for(long i = 0; i < size; i++)
    {
      pair < double *, long > *p = new pair < double *, long >;
      p->first = new double[dim];
      
      for(int j = 0; j < dim; j++)
	{
	  if(infile == 0)
	    die(__FILE__, __FUNCTION__,  __LINE__, "bad pattern");
	  
	  infile >> p->first[j]; 
	}
      
      if(infile == 0)
	die(__FILE__, __FUNCTION__,  __LINE__, "missing pattern class");
      
      infile >> p->second;
      
      labels.insert(p->second);
      
      if(labelCount.count(p->second) == 0)
	labelCount.insert(map < long, long > :: value_type(p->second, 1));
      else
	labelCount[p->second]++;
      
      _pointSet.push_back(p);
    }
  
  for(laiter = labels.begin(); laiter != labels.end(); laiter++)
    _labels.push_back(*laiter);

  /* Sort labels, then create label->index map. */
  
  sort(_labels.begin(), _labels.end());

  for(long i = 0; i < _labels.size(); i++)
    {
      _labelMap.insert(map < long, long > :: value_type(_labels[i], i));
      _labelCount.push_back(labelCount[_labels[i]]);

      cout << "dataset " << file << " label at index " << i << " is " << _labels[i] ;
      cout << " number of patterns of " << _labels[i] << " is " << _labelCount[i];
      cout << " (=" << getLabelCount(_labels[i]) << ")" << endl;
    }
  
  infile.close();

  _maxTries = (long) (log10(PFAIL) / log10((double) (size - 1) / size));
  
  cout << "maxTries " << _maxTries << " (PFAIL = " << PFAIL << ")" << endl;

  /* Build index. */
  
  for(int i = 0; i < numLabels(); i++)
    _index.push_back(new vector < long > );
  
  for(long i = 0; i < size; i++)
    _index[getLabelIndex(_pointSet[i]->second)]->push_back(i);

  for(int i = 0; i < numLabels(); i++)
    cout << "index " << i << " label " << _labels[i] << " _index[getLabelIndex(_labels[i])]->size() " << _index[getLabelIndex(_labels[i])]->size() << endl;
  
#ifdef MYDEBUG
  cout << "datasetEnv::datasetEnv pattern first";
  for(int i = 0; i < _dim; i++)
    cout << " " << *(_pointSet[0]->first + i);
  cout << " " << _pointSet[0]->second << endl;
  cout << "datasetEnv::datasetEnv pattern last";
  for(int i = 0; i < _dim; i++)
    cout << " " << *(_pointSet[size - 1]->first + i);
  cout << " " << _pointSet[size - 1]->second << endl;
  
  cout << "datasetEnv::datasetEnv size " << this->size() << " labels";
  for(int i = 0; i < numLabels(); i++)
    {
      cout << " " << getLabelIndex(getLabel(i)) << "/" << getLabel(i) << "/" << getLabelCount(getLabel(i));
      cout << "/" << _index[i]->size();
    }
  
  cout << endl;  
#endif

}

datasetEnv::~datasetEnv()
{
  for(long i = 0; i < _pointSet.size(); i++)
    {
      delete [] _pointSet[i]->first;
      delete _pointSet[i];
    }

  for(int i = 0; i < numLabels(); i++)
    delete _index[i];
}

double datasetEnv::act(point *p,
		       long action, 
		       vector < double > &state, 
		       bool &end)
{
  end = true;

  if(p->label() == getLabel(action))
    return 1;
  else 
    return 0;
}

double datasetEnv::test(team *tm,
			string prefix)
{
  /* Confusion matrix. */
  vector < long > cm;
  /* Dimension of confusion  matrix. */
  long matDim = numLabels() * numLabels();

  vector < pair < double *, long > * > :: iterator psiter, psiterend;
  vector < double > state;
  
  double score;
  
  cm.insert(cm.begin(), matDim, 0);

  state.reserve(_dim);
  
  long labelIndex;
  long action;
  
  long hit = 0;
  long mis = 0;

  long partfp;

  int i, j;

  /* Calculate confusion matrix. */

  for(psiter = _pointSet.begin(), psiterend = _pointSet.end();
      psiter != psiterend; psiter++)
    { 
      state.clear();
      
      /* Now, labelIndex can be matched with team action. */
      labelIndex = getLabelIndex((*psiter)->second);
      
      for(i = 0; i < _dim; i++)
	state.push_back(((*psiter)->first)[i]);
      
      action = tm->getAction(state);
      
      cm[labelIndex * numLabels() + action]++;
      
      if(labelIndex == action)
	{
	  /* Correct. */
	  hit++;
	}
      else
	{
	  /* Inorrect. */
	  mis++;
	}
    }

  /* Calculate score. */
  
  score = 0;

  for(i = 0; i < numLabels(); i++)
    score += (double) cm[i * numLabels() + i] / getLabelCount(getLabel(i));
  
  score = (double) score / numLabels();

  /* Print out results. */

  cout << "explicitEnv::test " << prefix << " " << _setName;
  cout << " hit " << hit << " mis " << mis << " rate " << (double) hit / (hit + mis);
  cout << " score " << score << " cm";
  for(i = 0; i < numLabels(); i++)
    cout << " " << getLabel(i);
  for(i = 0; i < matDim; i++)
    cout << " " << cm[i];
  cout << " dr fr";
  for(i = 0; i < numLabels(); i++)
    {
      cout << " " << (double) cm[i * numLabels() + i] / getLabelCount(getLabel(i));
      
      partfp = 0;
      for(j = 0; j < numLabels(); j++)
	if(i != j)
	  partfp += cm[j * numLabels() + i];

      cout << " " << (double) partfp / (size() - getLabelCount(getLabel(i)));
    }
  cout << " " << *tm << endl;

  return score;
}


