#ifndef learner_hpp
#define learner_hpp

#include <bitset>
#include <vector>
#include <cmath>
#include <map>

#include "misc.hpp"

#define REGISTERS 8
#define INPUTS 64

using namespace std;

typedef bitset < 6 + 3 + 3 + 1 > instruction;

class learner
{
private:
  
  /* Next id to use. */
  static long _count; 
  
  /* Unique id of learner. */
  unsigned long _id;
  /* Time step at which generated. */
  long _gtime; 
  /* Action index. */
  long _action;

  /* Number of references by teams. */
  int _nrefs;
  
  /* Bid program. */
  vector < instruction * > _bid;
  
  double run(vector < instruction * > &, double *, int, double *);
  
  /* Instruction definitions. */
  
  /* Masks. Apply before shifting. */
  
  static const instruction _modeMask;
  static const instruction _opMask;
  static const instruction _dstMask;
  static const instruction _srcMask;
  
  /* Modes. */
  
  static const instruction _mode0;
  static const instruction _mode1;
    
  /* Binary operations. */

  static const instruction _opSum;
  static const instruction _opDiff;
  static const instruction _opProd;
  static const instruction _opDiv;
  static const instruction _opMod;
  static const instruction _opCos;
  static const instruction _opExp;
  static const instruction _opLog;
  
  /* Shift amounts. */
  
  static const short _modeShift;
  static const short _opShift;
  static const short _dstShift;
  static const short _srcShift;
  
public:

  /* Create arbitrary learner. */
  learner(long, long, int);
  
  /* Create learner from another learner. */
  learner(long, learner &);
  
  ~learner();
  
  double bid(double *, int, double *);
  
  inline int size() 
  { return _bid.size(); }
  inline unsigned long id() 
  { return _id; }
  inline int action() 
  { return _action; }
  inline long gtime() 
  { return _gtime; }

  /* Mutate action. */
  inline bool muAction(long action) 
  { long a = _action; _action = action; return a != action; }
  /* Mutate bid. */
  bool muBid(double, double, double, double, int);

  inline int refInc() 
  { return ++_nrefs; }
  inline int refDec() 
  { return --_nrefs; }
  inline int refs() 
  { return _nrefs; }

  friend ostream & operator<<(ostream &, const learner &);
};


#endif
