#include "team.hpp"

long team::_count = 0;

team::~team()
{
  set < learner * > :: iterator leiter, leiterend;

  for(leiter = _members.begin(), leiterend = _members.end(); 
      leiter != leiterend; leiter++)
    (*leiter)->refDec();
}

void team::removeLearner(learner *lr)
{
  set < learner * > :: iterator leiter;
  
  if((leiter = _members.find(lr)) == _members.end())
    die(__FILE__, __FUNCTION__, __LINE__, "should not remove learner that is not there");
  
  _members.erase(leiter);
}

bool team::getOutcome(point *pt,
		      double *out)
{
  map < point *, double > :: iterator ouiter;
  
  if((ouiter = _outcomes.find(pt)) == _outcomes.end())
    return false;

  *out = ouiter->second;
  
  return true;
}

void team::setOutcome(point *pt,
		      double out)
{
  if((_outcomes.insert(map < point *, double >::value_type(pt, out))).second == false)
    die(__FILE__, __FUNCTION__, __LINE__, "could not set outcome, duplicate point?");
}

void team::deleteOutcome(point *pt)
{
  map < point *, double > :: iterator ouiter;
  
  if((ouiter = _outcomes.find(pt)) == _outcomes.end())
    die(__FILE__, __FUNCTION__, __LINE__, "should not delete outcome that is not set");

  _outcomes.erase(ouiter);
}

long team::getAction(vector < double > &state)
{
  set < learner * > :: iterator leiter, wiiter, leiterend;
  
  double maxBid;
  double nextBid;

  int ssize = state.size();
  
  double *REG = (double *) alloca(REGISTERS * sizeof(double));

  wiiter = leiter = _members.begin();
  maxBid = (*wiiter)->bid(&state[0], ssize, REG);

  for(leiter++, leiterend = _members.end(); 
      leiter != leiterend; leiter++)
    {
      nextBid = (*leiter)->bid(&state[0], ssize, REG);

      if(nextBid > maxBid)
	{
	  maxBid = nextBid;
	  wiiter = leiter;
	}
    }
  
  return (*wiiter)->action();
}

double team::distanceToSqrd(team *tm)
{
  double d = 0;
  
  map < point *, double > :: iterator myiter, ititer, myiterend;

  for(myiter = _outcomes.begin(), ititer = tm->_outcomes.begin(), myiterend = _outcomes.end();
      myiter != myiterend; myiter++, ititer++)
    {
      if(myiter->first != ititer->first)
	die(__FILE__, __FUNCTION__, __LINE__, "teams record different outcomes points");

      d += (myiter->second - ititer->second) * (myiter->second - ititer->second);
    }

  if(ititer != tm->_outcomes.end())
    die(__FILE__, __FUNCTION__, __LINE__, "teams record different outcomes numbers");
  
  return d;
}

void team::outcomes(vector < point * > &pts, 
		    vector < double > &oms)
{
  map < point *, double > :: iterator ouiter, ouiterend;
  
  for(ouiter = _outcomes.begin(), ouiterend = _outcomes.end(); 
      ouiter != ouiterend; ouiter++)
    {
      pts.push_back(ouiter->first);
      oms.push_back(ouiter->second);
    }
}

ostream & operator<<(ostream &os, 
		     const team &tm)
{
  set < learner * > :: iterator leiter, leiterend;

  os << "id " << tm._id << " size " << tm._members.size();
  os << " time " << tm._gtime;
  
  for(leiter = tm._members.begin(), leiterend = tm._members.end(); 
      leiter != leiterend; leiter++)
    os << " " << **leiter;
  
  return os;
}

